@@ -57,7 +57,8 @@ class MultiBodyConstraintSolver {
   SubmitProfileTiming profile_timing_func_{nullptr};
 
  public:
-  int pgs_iterations_{50};
+  bool keep_all_points_{false};
+  int pgs_iterations_{1};//increase if solver doesn't converge
   double least_squares_residual_threshold_{0};
   std::vector<int> limit_dependency_;
 
@@ -153,15 +154,48 @@ class MultiBodyConstraintSolver {
     return needs_outer_iterations_;
   }
 
+  virtual std::vector<ContactPoint> resolve_collision2(const std::vector<ContactPoint>& org_cps,
+                                                      const Scalar& dt) {
+    std::vector<ContactPoint> cps;
+    for (int i=0;i<org_cps.size();i++)
+    {
+      auto cp = org_cps[i];
+      if (keep_all_points_ || cp.distance<Algebra::zero())
+              cps.push_back(org_cps[i]);
+    }
+    resolve_collision_internal(cps, dt);
+    return cps;
+  }
+
+  virtual void resolve_collision(const std::vector<ContactPoint>& org_cps, const Scalar& dt) {
+    std::vector<ContactPoint> cps;
+    cps.resize(0);
+
+    for (int i=0;i<org_cps.size();i++)
+    {
+        auto cp = org_cps[i];
+        if (keep_all_points_ || cp.distance<Algebra::zero())
+            cps.push_back(org_cps[i]);
+    }
+    resolve_collision_internal(cps, dt);
+  }
+
+  
+
   // Solve impulse-based collision response using MLCP formulation from
   // Jakub Stępień, PhD Thesis, 2013.
   //
   // Args:
   // cps: contact points with distances < 0
   // dt : delta time (in seconds)
-  virtual void resolve_collision(const std::vector<ContactPoint>& cps,
-                                 const Scalar& dt) {
-    if (cps.empty()) return;
+
+  virtual void resolve_collision_internal(std::vector<ContactPoint> cps,
+                                const Scalar& dt) {
+  
+
+    if (cps.empty()) 
+        return;
+
     const int n_c = static_cast<int>(cps.size());
 
     const ContactPoint& cp0 = cps[0];
@@ -254,7 +288,7 @@ class MultiBodyConstraintSolver {
 
       const Vector3& world_point_a = cp.world_point_on_a;
       Matrix3X jac_a = point_jacobian2(*mb_a, cp.link_a, world_point_a, false);
-      VectorX jac_a_i = Algebra::mul_transpose(jac_a, cp.world_normal_on_b);
+      VectorX jac_a_i = Algebra::mul_transpose(jac_a, cp.world_normal_on_b*collision);
       Algebra::assign_horizontal(jac_con, jac_a_i, i, 0);
 
       const Vector3& world_point_b = cp.world_point_on_b;
@@ -264,7 +298,7 @@ class MultiBodyConstraintSolver {
       //     point_jacobian_fd(*mb_b, mb_b->m_q, cp.link_b,
       //     world_point_b);
       // jac_b.print("jac_b");
-      VectorX jac_b_i = Algebra::mul_transpose(jac_b, cp.world_normal_on_b);
+      VectorX jac_b_i = Algebra::mul_transpose(jac_b, cp.world_normal_on_b*collision);
       // jac_b_i.print("jac_b_i");
 
       std::vector<Scalar> qd_empty;
@@ -291,40 +325,39 @@ class MultiBodyConstraintSolver {
                  baumgarte_rel_vel;
       lcp_b[i] *= collision;
 
-#if USE_PROJECTED_FRICTION_DIRECTION
+//#define USE_PROJECTED_FRICTION_DIRECTION
+#ifdef USE_PROJECTED_FRICTION_DIRECTION
 
       // friction direction
       Vector3 lateral_rel_vel = rel_vel - normal_rel_vel * cp.world_normal_on_b;
       // lateral_rel_vel.print("lateral_rel_vel");
-      const Scalar lateral = Algebra::norm(lateral_rel_vel);
+      Scalar lateral = Algebra::norm(lateral_rel_vel);
       // printf("Algebra::norm(lateral_rel_vel): %.6f\n",
       //        Algebra::getDouble(lateral));
 
       Vector3 fr_direction1,fr_direction2;
       //      cp.world_normal_on_b.print("contact normal");
       //      fflush(stdout);
-      if(lateral < Algebra::fraction(1,10000)) {
-          // use the plane space of the contact normal as friction directions
-          plane_space(cp.world_normal_on_b,fr_direction1,fr_direction2);
-      }
-      else {
+
+      lateral = where_lt(lateral, Algebra::fraction(1,1000000),
+                                        Algebra::one(), lateral);
+
+      //if(lateral < Algebra::fraction(1,10000)) {
+      //    // use the plane space of the contact normal as friction directions
+      //    plane_space(cp.world_normal_on_b,fr_direction1,fr_direction2);
+      //}
+      //else 
+      {
           // use the negative lateral velocity and its orthogonal as friction
           // directions
           fr_direction1 = lateral_rel_vel * (Algebra::one() / lateral);
-          fr_direction1 = Algebra::normalize(fr_direction1);
-          fr_direction2 = Algebra::cross(fr_direction1,cp.world_normal_on_b);
-          fr_direction2 = Algebra::normalize(fr_direction2);
+          //fr_direction1 = Algebra::normalize(fr_direction1);
+          //fr_direction2 = Algebra::cross(fr_direction1,cp.world_normal_on_b);
+          //fr_direction2 = Algebra::normalize(fr_direction2);
       }
 #else
       // friction direction
-      Vector3 lateral_rel_vel = rel_vel - normal_rel_vel * cp.world_normal_on_b;
-      // if constexpr (is_cppad_scalar<Scalar>::value) {
-      if constexpr (true) {
-        // add epsilon to make prevent division by zero in gradient of norm
-        lateral_rel_vel[2] += Algebra::fraction(1, 100000);
-      }
-      const Scalar lateral = Algebra::norm(lateral_rel_vel);
-
+    
       Vector3 fr_direction1, fr_direction2;
       plane_space(cp.world_normal_on_b, fr_direction1, fr_direction2);
       fr_direction1 *= collision;
@@ -351,6 +384,8 @@ class MultiBodyConstraintSolver {
         VectorX jac_b_i_fr2 = Algebra::mul_transpose(jac_b, fr_direction2);
         Algebra::assign_horizontal(jac_con, jac_b_i_fr2, 2 * n_c + i, n_a);
       }
+      cps[i].fr_direction_1 = fr_direction1;
+      cps[i].fr_direction_2 = fr_direction2;
     }
 
     // jac_con.print("Constraint Jacobian");
@@ -407,7 +442,15 @@ class MultiBodyConstraintSolver {
                 least_squares_residual_threshold_, &con_lo, &con_hi);
       submit_profile_timing("");
     }
-    //    lcp_p.print("MLCP impulse solution");
+    // Algebra::print("MLCP", lcp_p);
+
+    // Save forces in contact points
+    for (int i = 0; i < n_c; ++i) {
+      ContactPoint& cp = cps[i];
+      cp.normal_force = lcp_p[i * 3 + 0]/dt;
+      cp.lateral_friction_force_1 = lcp_p[i * 3 + 1]/dt;
+      cp.lateral_friction_force_2 = lcp_p[i * 3 + 2]/dt;
+    }
 
     if (n_a > 0) {
       //normal impulse
@@ -514,4 +557,4 @@ static TINY_INLINE MultiBodyConstraintSolver<AlgebraTo> clone(
     const MultiBodyConstraintSolver<AlgebraFrom>& s) {
   return s.template clone<AlgebraTo>();
 }
-}  // namespace tds
\ No newline at end of file
+}  // namespace tds
