@@ -33,15 +33,22 @@ class MultiBody {
   typedef std::vector<Link> LinkCollection;
 
   /**
-   * Number of degrees of freedom, excluding floating-base coordinates.
+   * Number of positional degrees of freedom, excluding floating-base
+   * coordinates.
    */
-  int dof_{0};
+  int dof_q_{0};
 
   /**
-   * A container to store the nr of spherical joints to account for the discrepancy between the size of the coordinate
-   * vector and the nr of DoF
+   * Number of velocity/acceleration degrees of freedom, excluding floating-base
+   * coordinates.
    */
-  int spherical_joints_{0};
+  int dof_qd_{0};
+
+  /**
+   * Damping factor, applied to spherical joints, see integrator.hpp
+   */
+
+  Scalar joint_damping_{0.995};
 
   /**
    * Whether this system is floating or fixed to the world frame.
@@ -67,7 +74,7 @@ class MultiBody {
 
   std::vector<int> visual_shape_uids_;
   std::vector<int> visual_instance_uids_;
-  
+
   // offset of geometry (relative to the base frame)
   std::vector<Transform> X_visuals_;
 
@@ -75,13 +82,13 @@ class MultiBody {
   // offset of collision geometries (relative to this link frame)
   std::vector<Transform> X_collisions_;
 
-public:
+ public:
   VectorX q_, qd_, qdd_, tau_;
 
   std::string name_;
 
   LinkCollection links_;
- 
+
   explicit MultiBody(bool isFloating = false) : is_floating_(isFloating) {}
 
   template <typename AlgebraTo = Algebra>
@@ -89,11 +96,14 @@ public:
     typedef Conversion<Algebra, AlgebraTo> C;
     MultiBody<AlgebraTo> conv(is_floating_);
     conv.name_ = name_;
-    conv.dof_ = dof_;
+    conv.dof_q_ = dof_q_;
+    conv.dof_qd_ = dof_qd_;
     conv.q_ = C::convert(q_);
     conv.qd_ = C::convert(qd_);
     conv.qdd_ = C::convert(qdd_);
     conv.tau_ = C::convert(tau_);
+    conv.joint_damping_ = C::convert(joint_damping_);
+    conv.control_indices_ = control_indices_;
     for (const auto &link : links_) {
       conv.links_.push_back(link.template clone<AlgebraTo>());
     }
@@ -120,7 +130,7 @@ public:
   TINY_INLINE const LinkCollection &links() const { return links_; }
   TINY_INLINE std::size_t num_links() const { return links_.size(); }
   TINY_INLINE std::size_t size() const { return links_.size(); }
-  
+
   TINY_INLINE const Link &operator[](std::size_t i) const { return links_[i]; }
   TINY_INLINE Link &operator[](std::size_t i) { return links_[i]; }
   TINY_INLINE typename LinkCollection::iterator begin() {
@@ -141,21 +151,18 @@ public:
   }
   TINY_INLINE bool empty() const { return links_.empty(); }
 
-  /**
-   * Return the number of spherical joints in the system (to account for the difference between DoF and length of q_
-   */
-  TINY_INLINE int spherical_joints() const {return spherical_joints_;}
-
   /**
    * Dimensionality of joint positions q (including 7-DoF floating-base
    * coordinates if this system is floating-base).
    */
-  TINY_INLINE int dof() const { return is_floating_ ? dof_ + 7 : dof_; }
+  TINY_INLINE int dof() const { return is_floating_ ? dof_q_ + 7 : dof_q_; }
   /**
    * Dimensionality of joint velocities qd and accelerations qdd (including
    * 6-DoF base velocity and acceleration, if this system is floating-base).
    */
-  TINY_INLINE int dof_qd() const { return is_floating_ ? dof_ + 6 : dof_; }
+  TINY_INLINE int dof_qd() const {
+    return is_floating_ ? dof_qd_ + 6 : dof_qd_;
+  }
 
   /**
    * Dimensionality of control input, i.e. number of actuated DOFs.
@@ -216,18 +223,22 @@ public:
   TINY_INLINE const ArticulatedBodyInertia &base_abi() const {
     return base_abi_;
   }
-  TINY_INLINE Transform &base_X_world() { 
-      return base_X_world_; 
+  TINY_INLINE Transform &base_X_world() { return base_X_world_; }
+  TINY_INLINE const Transform &base_X_world() const { return base_X_world_; }
+
+  TINY_INLINE std::vector<int> &visual_shape_ids() {
+    return visual_shape_uids_;
   }
-  TINY_INLINE const Transform &base_X_world() const { 
-      return base_X_world_; 
+  TINY_INLINE const std::vector<int> &visual_shape_ids() const {
+    return visual_shape_uids_;
   }
 
-  TINY_INLINE std::vector<int> &visual_shape_ids() { return visual_shape_uids_; }
-  TINY_INLINE const std::vector<int> &visual_shape_ids() const { return visual_shape_uids_; }
-
-  TINY_INLINE std::vector<int>& visual_instance_uids() { return visual_instance_uids_; }
-  TINY_INLINE const std::vector<int>& visual_instance_uids() const { return visual_instance_uids_; }
+  TINY_INLINE std::vector<int> &visual_instance_uids() {
+    return visual_instance_uids_;
+  }
+  TINY_INLINE const std::vector<int> &visual_instance_uids() const {
+    return visual_instance_uids_;
+  }
 
   TINY_INLINE std::vector<Transform> &X_visuals() { return X_visuals_; }
   TINY_INLINE const std::vector<Transform> &X_visuals() const {
@@ -278,24 +289,30 @@ public:
     }
   }
 
-  void set_orientation(const Quaternion& initial_orientation) {
-      base_X_world_.rotation = Algebra::quat_to_matrix(initial_orientation);
-      if (is_floating_) {
-          q_[0] = Algebra::quat_x(initial_orientation);
-          q_[1] = Algebra::quat_y(initial_orientation);
-          q_[2] = Algebra::quat_z(initial_orientation);
-          q_[3] = Algebra::quat_w(initial_orientation);
-      }
+  void set_orientation(const Quaternion &initial_orientation) {
+    base_X_world_.rotation = Algebra::quat_to_matrix(initial_orientation);
+    if (is_floating_) {
+      q_[0] = Algebra::quat_x(initial_orientation);
+      q_[1] = Algebra::quat_y(initial_orientation);
+      q_[2] = Algebra::quat_z(initial_orientation);
+      q_[3] = Algebra::quat_w(initial_orientation);
+    }
+  }
+  void set_orientation(const Matrix3 &rotation) {
+    base_X_world_.rotation = rotation;
+    if (is_floating_) {
+      auto quat = Algebra::matrix_to_quat(rotation);
+      q_[0] = Algebra::quat_x(quat);
+      q_[1] = Algebra::quat_y(quat);
+      q_[2] = Algebra::quat_z(quat);
+      q_[3] = Algebra::quat_w(quat);
+    }
   }
 
-  Vector3 get_position() const
-  {
-      return get_world_transform(-1).translation;
-  }
+  Vector3 get_position() const { return get_world_transform(-1).translation; }
 
-  Quaternion get_orientation() const
-  {
-      return Algebra::matrix_to_quat(get_world_transform(-1).rotation);
+  Quaternion get_orientation() const {
+    return Algebra::matrix_to_quat(get_world_transform(-1).rotation);
   }
 
   /**
@@ -306,21 +323,23 @@ public:
     // make sure dof and the q / qd indices in the links are accurate
     int q_index = is_floating_ ? 7 : 0;
     int qd_index = is_floating_ ? 6 : 0;
-    dof_ = 0;  // excludes floating-base DOF
-    spherical_joints_ = 0;
+    dof_q_ = 0;   // excludes floating-base DOF
+    dof_qd_ = 0;  // excludes floating-base DOF
+
     for (Link &link : links_) {
       assert(link.index >= 0);
       link.q_index = q_index;
       link.qd_index = qd_index;
       if (link.joint_type == JOINT_SPHERICAL) {
-          q_index += 4;
-          qd_index += 3;
-          dof_ += 3;
-          ++spherical_joints_;
-      } else if(link.joint_type != JOINT_FIXED) {
+        q_index += 4;
+        qd_index += 3;
+        dof_q_ += 4;
+        dof_qd_ += 3;
+      } else if (link.joint_type != JOINT_FIXED) {
         ++q_index;
         ++qd_index;
-        ++dof_;
+        ++dof_q_;
+        ++dof_qd_;
       } else {
         link.q_index = -2;
         link.qd_index = -2;
@@ -330,9 +349,9 @@ public:
     q_ = Algebra::zerox(q_index);
     // Initialize the quaternions for spherical joints
     for (Link &link : links_) {
-        if (link.joint_type == JOINT_SPHERICAL) {
-            q_[link.q_index + 3] = Algebra::one();
-        }
+      if (link.joint_type == JOINT_SPHERICAL) {
+        q_[link.q_index + 3] = Algebra::one();
+      }
     }
     qd_ = Algebra::zerox(dof_qd());
     qdd_ = Algebra::zerox(dof_qd());
@@ -341,14 +360,16 @@ public:
       q_[3] = Algebra::one();  // make sure orientation is valid
     }
 
+    clear_forces();
+    base_acceleration_.set_zero();
     base_abi_ = base_rbi_;
 
     if (is_floating_ && !base_abi_.is_invertible()) {
-      
       fprintf(stderr,
               "Error: floating-base inertia matrix (ABI) is not invertible. "
               "Are you sure the model should be floating-base?\n");
-      // requires to_double, which breaks Algebra::print("Floating-base ABI", base_abi_);
+      // requires to_double, which breaks Algebra::print("Floating-base ABI",
+      // base_abi_);
       assert(0);
       exit(1);
     }
@@ -360,7 +381,8 @@ public:
   template <typename Algebra2>
   MultiBody(const MultiBody<Algebra2> &mb)
       : links_(mb.links_),
-        dof_(mb.dof_),
+        dof_q_(mb.dof_q_),
+        dof_qd_(mb.dof_qd_),
         control_indices_(mb.control_indices_),
         is_floating_(mb.is_floating_),
         base_velocity_(mb.base_velocity_),
@@ -427,16 +449,14 @@ public:
   /**
    * Transforms a point in body coordinates to world coordinates.
    */
-  inline Vector3 body_to_world(int link_index,
-      const Vector3& point) const {
-      return get_world_transform(link_index).apply(point);
+  inline Vector3 body_to_world(int link_index, const Vector3 &point) const {
+    return get_world_transform(link_index).apply(point);
   }
   /**
    * Transforms a point in world coordinates to bodyt coordinates.
    */
-  inline Vector3 world_to_body(int link_index,
-      const Vector3& point) const {
-      return get_world_transform(link_index).apply_inverse(point);
+  inline Vector3 world_to_body(int link_index, const Vector3 &point) const {
+    return get_world_transform(link_index).apply_inverse(point);
   }
 
   /**
@@ -449,98 +469,106 @@ public:
     if (link == -1) {
       return tf.apply(base_rbi_.com);
     } else {
-      return tf.apply(links_[link].I.com);
+      return tf.apply(links_[link].rbi.com);
     }
   }
 
-  TINY_INLINE void set_q(const VectorX& q)
-  {
-      q_ = q;
+  TINY_INLINE void set_q(const VectorX &q) { q_ = q; }
+
+  //  TINY_INLINE Scalar get_q_for_link(const VectorX &q, int link_index) const
+  //  {
+  //    if (Algebra::size(q) == 0) return Algebra::zero();
+  //    const Link &link = links_[link_index];
+  //    return link.joint_type == JOINT_FIXED ? Algebra::zero() :
+  //    q[link.q_index];
+  //  }
+  //  TINY_INLINE Scalar get_q_for_link(int link_index) const {
+  //    get_q_for_link(q_, link_index);
+  //  }
+  TINY_INLINE VectorX get_q_for_link(const VectorX &q, int link_index) const {
+    const Link &link = links_[link_index];
+
+    if (Algebra::size(q) == 0 || link.joint_type == JOINT_FIXED) {
+      return link.joint_type == JOINT_SPHERICAL ? Algebra::zerox(4)
+                                                : Algebra::zerox(1);
+    }
+
+    return link.joint_type == JOINT_SPHERICAL ? q.segment(link.q_index, 4)
+                                              : q.segment(link.q_index, 1);
+  }
+  TINY_INLINE VectorX get_q_for_link(int link_index) const {
+    get_q_for_link(q_, link_index);
   }
 
-//  TINY_INLINE Scalar get_q_for_link(const VectorX &q, int link_index) const {
-//    if (Algebra::size(q) == 0) return Algebra::zero();
-//    const Link &link = links_[link_index];
-//    return link.joint_type == JOINT_FIXED ? Algebra::zero() : q[link.q_index];
-//  }
-//  TINY_INLINE Scalar get_q_for_link(int link_index) const {
-//    get_q_for_link(q_, link_index);
-//  }
-    TINY_INLINE VectorX get_q_for_link(const VectorX &q, int link_index) const {
-        const Link &link = links_[link_index];
+  //  TINY_INLINE Scalar get_qd_for_link(const VectorX &qd, int link_index)
+  //  const {
+  //    if (Algebra::size(qd) == 0) return Algebra::zero();
+  //    const Link &link = links_[link_index];
+  //    return link.joint_type == JOINT_FIXED ? Algebra::zero() :
+  //    qd[link.qd_index];
+  //  }
+  //  TINY_INLINE Scalar get_qd_for_link(int link_index) const {
+  //    return get_qd_for_link(qd_, link_index);
+  //  }
 
-        if (Algebra::size(q) == 0 || link.joint_type == JOINT_FIXED){
-            return link.joint_type == JOINT_SPHERICAL ? Algebra::zerox(4) : Algebra::zerox(1);
-        }
+  TINY_INLINE VectorX get_qd_for_link(const VectorX &qd, int link_index) const {
+    const Link &link = links_[link_index];
 
-        return link.joint_type == JOINT_SPHERICAL ? q.segment(link.q_index, 4) : q.segment(link.q_index, 1);
-    }
-    TINY_INLINE VectorX get_q_for_link(int link_index) const {
-        get_q_for_link(q_, link_index);
+    if (Algebra::size(qd) == 0 || link.joint_type == JOINT_FIXED) {
+      return link.joint_type == JOINT_SPHERICAL ? Algebra::zerox(3)
+                                                : Algebra::zerox(1);
     }
 
-//  TINY_INLINE Scalar get_qd_for_link(const VectorX &qd, int link_index) const {
-//    if (Algebra::size(qd) == 0) return Algebra::zero();
-//    const Link &link = links_[link_index];
-//    return link.joint_type == JOINT_FIXED ? Algebra::zero() : qd[link.qd_index];
-//  }
-//  TINY_INLINE Scalar get_qd_for_link(int link_index) const {
-//    return get_qd_for_link(qd_, link_index);
-//  }
+    return link.joint_type == JOINT_SPHERICAL ? qd.segment(link.qd_index, 3)
+                                              : qd.segment(link.qd_index, 1);
+  }
+  TINY_INLINE VectorX get_qd_for_link(int link_index) const {
+    return get_qd_for_link(qd_, link_index);
+  }
 
-    TINY_INLINE VectorX get_qd_for_link(const VectorX &qd, int link_index) const {
-        const Link &link = links_[link_index];
+  TINY_INLINE VectorX get_qdd_for_link(const VectorX &qdd,
+                                       int link_index) const {
+    return get_qd_for_link(qdd, link_index);
+  }
+  TINY_INLINE VectorX get_qdd_for_link(int link_index) const {
+    return get_qdd_for_link(qdd_, link_index);
+  }
+  //  TINY_INLINE Scalar get_qdd_for_link(const VectorX &qdd,
+  //                                      int link_index) const {
+  //    return get_qd_for_link(qdd, link_index);
+  //  }
+  //  TINY_INLINE Scalar get_qdd_for_link(int link_index) const {
+  //    return get_qdd_for_link(qdd_, link_index);
+  //  }
 
-        if (Algebra::size(qd) == 0 || link.joint_type == JOINT_FIXED){
-            return link.joint_type == JOINT_SPHERICAL ? Algebra::zerox(3) : Algebra::zerox(1);
-        }
+  //  TINY_INLINE Scalar get_tau_for_link(const VectorX &tau,
+  //                                      int link_index) const {
+  //    if (Algebra::size(tau) == 0) return Algebra::zero();
+  //    const Link &link = links_[link_index];
+  //    int offset = is_floating_ ? -6 : 0;
+  //    return link.joint_type == JOINT_FIXED ? Algebra::zero()
+  //                                          : tau[link.qd_index + offset];
+  //  }
+  //  TINY_INLINE Scalar get_tau_for_link(int link_index) const {
+  //    return get_tau_for_link(tau_, link_index);
+  //  }
+  TINY_INLINE VectorX get_tau_for_link(const VectorX &tau,
+                                       int link_index) const {
+    const Link &link = links_[link_index];
 
-        return link.joint_type == JOINT_SPHERICAL ? qd.segment(link.qd_index, 3) : qd.segment(link.qd_index, 1);
-    }
-    TINY_INLINE VectorX get_qd_for_link(int link_index) const {
-        return get_qd_for_link(qd_, link_index);
+    if (Algebra::size(tau) == 0 || link.joint_type == JOINT_FIXED) {
+      return link.joint_type == JOINT_SPHERICAL ? Algebra::zerox(3)
+                                                : Algebra::zerox(1);
     }
+    int offset = is_floating_ ? -6 : 0;
 
-    TINY_INLINE VectorX get_qdd_for_link(const VectorX &qdd,
-                                        int link_index) const {
-        return get_qd_for_link(qdd, link_index);
-    }
-    TINY_INLINE VectorX get_qdd_for_link(int link_index) const {
-        return get_qdd_for_link(qdd_, link_index);
-    }
-//  TINY_INLINE Scalar get_qdd_for_link(const VectorX &qdd,
-//                                      int link_index) const {
-//    return get_qd_for_link(qdd, link_index);
-//  }
-//  TINY_INLINE Scalar get_qdd_for_link(int link_index) const {
-//    return get_qdd_for_link(qdd_, link_index);
-//  }
-
-//  TINY_INLINE Scalar get_tau_for_link(const VectorX &tau,
-//                                      int link_index) const {
-//    if (Algebra::size(tau) == 0) return Algebra::zero();
-//    const Link &link = links_[link_index];
-//    int offset = is_floating_ ? -6 : 0;
-//    return link.joint_type == JOINT_FIXED ? Algebra::zero()
-//                                          : tau[link.qd_index + offset];
-//  }
-//  TINY_INLINE Scalar get_tau_for_link(int link_index) const {
-//    return get_tau_for_link(tau_, link_index);
-//  }
-    TINY_INLINE VectorX get_tau_for_link(const VectorX &tau,
-                                        int link_index) const {
-        const Link &link = links_[link_index];
-
-        if (Algebra::size(tau) == 0 || link.joint_type == JOINT_FIXED){
-            return link.joint_type == JOINT_SPHERICAL ? Algebra::zerox(3) : Algebra::zerox(1);
-        }
-        int offset = is_floating_ ? -6 : 0;
-
-        return link.joint_type == JOINT_SPHERICAL ? tau.segment(link.qd_index + offset, 3) : tau.segment(link.qd_index + offset, 1);
-    }
-    TINY_INLINE VectorX  get_tau_for_link(int link_index) const {
-        return get_tau_for_link(tau_, link_index);
-    }
+    return link.joint_type == JOINT_SPHERICAL
+               ? tau.segment(link.qd_index + offset, 3)
+               : tau.segment(link.qd_index + offset, 1);
+  }
+  TINY_INLINE VectorX get_tau_for_link(int link_index) const {
+    return get_tau_for_link(tau_, link_index);
+  }
 
   /**
    * Set joint torques and external forces in all links and the base to zero.
@@ -570,10 +598,11 @@ public:
     if (!links_.empty()) parent_index = static_cast<int>(links_.size()) - 1;
     return attach(link, parent_index, is_controllable);
   }
- 
-  size_t attach_link(Link &link, int parent_index, bool is_controllable = true) {
-     return attach(link, parent_index, is_controllable); 
- } 
+
+  size_t attach_link(Link &link, int parent_index,
+                     bool is_controllable = true) {
+    return attach(link, parent_index, is_controllable);
+  }
   size_t attach(Link &link, int parent_index, bool is_controllable = true) {
     int sz = static_cast<int>(links_.size());
     assert(parent_index < sz);
@@ -584,8 +613,8 @@ public:
       // assert(Algebra::norm(link.S) > Algebra::zero());
       link.q_index = dof();
       link.qd_index = dof_qd();
-      dof_ += 3;
-      ++spherical_joints_;
+      dof_q_ += 4;
+      dof_qd_ += 3;
       // not sure about this:
       if (is_controllable) {
         if (control_indices_.empty()) {
@@ -598,13 +627,15 @@ public:
           control_indices_.push_back(control_indices_.back() + 1);
         }
       }
-    }else if (link.joint_type != JOINT_FIXED) {
+    } else if (link.joint_type != JOINT_FIXED) {
       assert(Algebra::norm(link.S) > Algebra::zero());
-      // Is this redundant? Or can a multibody be created used without a call to initialize()? If redundant, delete?
-      // Else, how to fix now that dof != length q_?
+      // Is this redundant? Or can a multibody be created without a call to
+      // initialize()? If redundant, delete? Else, how to fix now that dof !=
+      // length q_?
       link.q_index = dof();
       link.qd_index = dof_qd();
-      dof_++;
+      dof_q_++;
+      dof_qd_++;
       if (is_controllable) {
         if (control_indices_.empty()) {
           control_indices_.push_back(0);
@@ -625,8 +656,13 @@ public:
 //    link.S.print("joint.S");
 #endif
     links_.push_back(link);
-    return links_.size()-1;
+    return links_.size() - 1;
   }
+
+  void set_joint_damping(Scalar damping) {
+    joint_damping_ = Algebra::clamp(damping, 0, 1);
+  }
+  Scalar joint_damping() const { return joint_damping_; }
 };
 
 template <typename AlgebraFrom, typename AlgebraTo = AlgebraFrom>
@@ -635,4 +671,4 @@ static TINY_INLINE MultiBody<AlgebraTo> clone(
   return mb.template clone<AlgebraTo>();
 }
 }  // namespace tds
-#endif //_MULTI_BODY_HPP
+#endif  //_MULTI_BODY_HPP
